import os.path
import sys
from tkinter import *
from tkinter import ttk, StringVar, simpledialog, messagebox, filedialog
from com.canberra.datasources.spectroscopy.cam import Datasource, ParameterCodes
from com.canberra.datasources.spectroscopy.cam import OpenOptions, CloseOptions
from com.canberra.nuclides import Nuclide, Line
from com.canberra.spectroscopy.peaks import PeakAttributes
from com.canberra.units.activity import Custom
from com.canberra.exceptions import NativeErrorException
import traceback



class LabLists:
    basefile = "C:/GENIE2K/SCRIPTS/LabLists.txt"

    def __init__(self, fname=basefile, columns=None):
        self._quantity = []
        self._resultUnit = []
        self._nuclideType = []
        self._AnalysisMethodology = []
        self._Laboratories = []
        self._LabQualifiers = ["Approved", "Estimated", "Less Than Lc", "Rejected"]

        if columns is None:
            columns = ["quantity", "resultUnit", "nuclideType", "AnalysisMethodology", "Laboratories",
                       "LabQualifiers"]
            if os.path.exists(fname):
                with open(fname, 'r', encoding='utf-8-sig') as filelist:
                    firstline = True
                    coldict = {"quantity": 0, "resultUnit": 1, "nuclideType": 2, "AnalysisMethodology": 3,
                               "Laboratories": 4, "LabQualifiers": 5}
                    for line in filelist.readlines():
                        ls = line.replace("\n", "").split(',')
                        if firstline:
                            tmp = {}
                            allcolumnsfound = True
                            for key in coldict.keys():
                                if key in ls:
                                    tmp[key] = ls.index(key)
                                else:
                                    allcolumnsfound = False
                                    break
                            if allcolumnsfound:
                                coldict = tmp
                            firstline = False
                        else:
                            self._quantity.append(ls[coldict["quantity"]])
                            self._resultUnit.append(ls[coldict["resultUnit"]])
                            self._nuclideType.append(ls[coldict["nuclideType"]][:64])
                            self._AnalysisMethodology.append(ls[coldict["AnalysisMethodology"]][:64])
                            self._Laboratories.append(ls[coldict["Laboratories"]][:64])
                            # self._LabQualifiers.append(ls[coldict["LabQualifiers"]])

    @property
    def quantity(self):
        return list(filter(None, self._quantity))

    @property
    def resultUnit(self):
        return list(filter(None, self._resultUnit))

    @property
    def nuclideType(self):
        return list(filter(None, self._nuclideType))

    @property
    def AnalysisMethodology(self):
        return list(filter(None, self._AnalysisMethodology))

    @property
    def Laboratories(self):
        return list(filter(None, self._Laboratories))

    @property
    def LabQualifiers(self):
        return list(filter(None, self._LabQualifiers))


class DisplayCustomPrompt:

    def __init__(self, ds: Datasource):
        tmpfname = "C:/Genie2k/VDMTEMP/tmpfile.cnf"
        self.ds = ds
        self._criticalLevel = {}
        self._skipMDA = bool()
        ds.saveAs(name=tmpfname, overwrite=True)
        self._nucIdList = list()
        self._nucIndex = list()
        with Datasource() as tmp:
            tmp.open(tmpfname, OpenOptions.READWRITE)
            # self._nucIdList = [nuc for nuc in tmp.getNuclides(fullNidResults=True) if nuc.identified]
            # self._nucIndex = [nuc.name for nuc in tmp.nuclides]
            for nuc in tmp.getNuclides(fullNidResults=True):
                self._nucIndex.append(nuc.name)
                if nuc.identified:
                    self._nucIdList.append(nuc)
        # self._nucIndex = [nuc.name for nuc in hldr]
        if os.path.isfile(tmpfname):
            os.remove(tmpfname)
        self._lablist = LabLists()
        self._show()

    def _pushtocam(self):
        """
        internal function which writes the data to the cam datasource
        note: this will save the cam file/detector.
        :return: None
        """
        try:
            # check to make sure all fields are populated
            if self._verifyinputs():
                self._saveparams()
                self.ds.save()
                self._wrtieoutput()
                self._close()
            else:
                messagebox.showinfo("Save Failed", "Not all fields have been populated\n"
                                                   "Please populate all fields to continue")
        except Exception as exception:
            messagebox.showerror("Error in python script", traceback.format_exc())
            raise exception

    def _verifyinputs(self):
        """
        Checks each input variable to see if it is populated
        Note: this does not check to see if every ID'ed nuclide has Lab Qualifier which was manually set.
            Lab Qualifiers are automatically set to either "Estimated" or "Less than Lc" based on criteria
                See onLoad for details

        :return: True if everything is populated/has selected value
        """
        return self._outfile.get() != "" and self._reportedBy.get() != "" and self._idBarCode.get() != "" and \
            self._qcBatchNum.get() != "" and self._analysisRequest.get() != "" and self._cb_method.get() != "" and \
            self._cb_wetdry.get() != "" and self._cb_labname.get() != ""

    def _wrtieoutput(self):
        """
        Gathers together all the data for each line and writes it line by line to the file specified.
        If the file does not exist, it will be created.  If the directory does not exist, it is created.
        :return:
        """
        # get the file name
        fileName = str(self._outfile.get())
        # check if the directory exists, if not, try and create the directory.
        if not os.path.exists(os.path.split(fileName)[0]):
            os.mkdir(os.path.split(fileName)[0])

        # The first seven columns of the output are repeated for all lines. This holds those values
        lineHeader = [str(self.ds.getParameter(ParameterCodes.CAM_X_STIME)), str(self._reportedBy.get()),
                      str(self._idBarCode.get()), str(self._qcBatchNum.get()), str(self._analysisRequest.get()),
                      str(self._cb_labname.get()), str(self._cb_method.get())]

        # this statement opens the file 'a' means append, which will either create a new file or append an existing file
        with open(fileName, 'a') as outputFile:
            # If not appending, then write header line
            if os.stat(fileName).st_size == 0:
                columnheader = "Analytical Result Import\n" \
                               "Result Date*, Reported By*, ID/Barcode*, QC Batch #, Analysis Request Name, " \
                               "Laboratory Name, Analysis Methodology*, Nuclide Type*, Result*, Result Unit*, " \
                               "Uncertainty/Error, Coverage Factor, MDA/MDC, Measured Critical Level, " \
                               "Quantity as Analyzed, Quantity Unit, Wet or Dry?, Lab Qualifier, Comment, Upload Type\n"
                outputFile.write(columnheader)

            # get the error multiplier value
            numsigma = self.ds.getParameter(ParameterCodes.CAM_F_SIGMA)
            # get the activity multiplier value
            multiact = self.ds.getParameter(ParameterCodes.CAM_F_ACTIVMULT)
            sampsize = self.ds.getParameter(ParameterCodes.CAM_F_SQUANT)
            sampUnit = self.ds.getParameter(ParameterCodes.CAM_T_SUNITS)

            # variable to hold all the nuclide data lines
            lineMiddle = []

            # below is a for loop which will loop through list of id'ed nuclides from the cam file
            for nuclide in self._nucIdList:
                # get the cam record number of the nuclide.
                # this is used to get the custom cam parameter holding the Lab Qualifier value
                recnum = self._nucIndex.index(nuclide.name) + 1
                # lab qualifier value (1-4)  approved, estiamted, less than lc, rejected
                lbqual = self.ds.getParameter(ParameterCodes.CAM_F_NCLSP1, recnum)

                # hold activity/uncert values
                actval = nuclide.activity.value
                actunc = nuclide.activity.uncertainty
                # critical level
                critlvl = ""
                if nuclide.properties['mdaComputed']:
                    critlvl = self._criticalLevel[nuclide.name]
                if self._whatactreport.get() == 2:
                    # get the key line if more than one key line then return weighted mean
                    if nuclide.properties['keyLineEntryNo'] > 0:
                        actval = nuclide.lines[nuclide.properties['keyLineEntryNo'] - 1].activity.value
                        actunc = nuclide.lines[nuclide.properties['keyLineEntryNo'] - 1].activity.uncertainty
                # add activity, unit, act uncert, error multiplier, mda, crit level to middle
                lineMiddle.append(
                    [nuclide.name, str(actval), f'{str(nuclide.activity.units.label)}/{str(sampUnit)}', str(actunc),
                     str(numsigma), str(nuclide.mda.value), str(critlvl), str(sampsize), str(sampUnit),
                     'dry' if self._whatactreport.get() == 1 else 'wet',
                     str(self._lablist.LabQualifiers[int(lbqual) - 1])])
            for midline in lineMiddle:
                # merge the lists and write them out as a csv line (ending in \n for new line)
                outputFile.write(','.join(lineHeader) + ',' + ','.join(midline) + '\n')

    def _onLoad(self):
        """
        Checks if any MDA step has been run and if any lab qualifiers have been set
        If MDA has been run this fills the _criticalLevel dictionary
        If no lob qualifier has been set for a particular nuclide, then set
        to either "Estimated" or "Less than Lc" based on evaluation.

        Called on initial loading of window. (from "_show" method)

        :return: None
        """
        self._skipMDA = False
        try:
            # this will throw an error if no mda step has ever been run
            # which is why this is in a try - except

            # useISO11929 is just a convenient means to check if mda has been run
            # and check which critical level is to be used.
            useISO11929 = self.ds.getParameter(ParameterCodes.CAM_L_NID11929DONE)
        except:
            # no detection limit step has been done so flag
            useISO11929 = False
            self._skipMDA = True

        # useISO11929 is needed because the standard mda step will
        # populate the DecisionLevel value though it will be in channels.
        # So if the standard mda was done we have to divide mda by Detection limit
        # and multiply by decision level
        try:
            # grab the activity mulitplier
            multiact = self.ds.getParameter(ParameterCodes.CAM_F_ACTIVMULT)

            # loop through id'ed nuclides to check mda/critical level and
            # check/set initial Lab Qualifier value
            for nuclide in self._nucIdList:
                # hold record number of nuclide for later use in setParameter
                recnum = self._nucIndex.index(nuclide.name) + 1
                # get the cam parameter holding the lab qualifier number
                labQualValue = self.ds.getParameter(ParameterCodes.CAM_F_NCLSP1, recnum)
                # set the flag for less than critical level to false at start of each loop
                lessthanLc = False
                # set the critical level value to the largest float at the start of each loop.
                critlvl = sys.float_info.max
                if not self._skipMDA:
                    if nuclide.properties['mdaComputed']:
                        if useISO11929:
                            critlvl = nuclide.properties['11929DecisionLevel'] * multiact
                        else:
                            # loop through all energy lines for each nuclide
                            for l in nuclide.lines:
                                if l.properties['detectionLimit'] != 0:
                                    # The "nuclide" critical level is the lowest critlvl of all nuclide lines.
                                    # So, the critlvl variable is set to the largest possible value
                                    # then compared to every line value.
                                    # if the current line value is less than critlvl,
                                    # then crtilvl is set equal to the current line value.
                                    # Repeat this process for all lines in a nuclide to get the lowest critlvl value.
                                    currentCritLvl = l.mda.value * (l.properties['decisionLevel'] /
                                                                    l.properties['detectionLimit'])
                                    if currentCritLvl < critlvl:
                                        critlvl = currentCritLvl
                        # add the nuclide name as the key and critlvl as the value to the dictionary
                        self._criticalLevel[nuclide.name] = critlvl
                        # check if activity is less than the critial level
                        lessthanLc = nuclide.activity.value < critlvl
                if labQualValue < 1:
                    # if labQualValue is 0 then it had not been set to set the default value of 2 : "Estimated"
                    self.ds.setParameter(ParameterCodes.CAM_F_NCLSP1, 2, recnum)
                    if lessthanLc:
                        # if lessthanLc == true then set the value to 3 : "Less than Lc"
                        self.ds.setParameter(ParameterCodes.CAM_F_NCLSP1, 3, recnum)
        except Exception as exception:
            messagebox.showerror("Error In Script", traceback.format_exc())
            raise exception

    def _loadparams(self):
        """
        Read all cam params for fields on the primary window
        :return:None
        """
        try:
            dsname = str(self.ds.name).replace("DET:", "")
            if os.path.isfile(dsname):
                dsname = os.path.split(dsname)[1][:-4]

            self._outfile.set("C:/GENIE2K/CustomOutput/" + dsname + ".csv")
            self._reportedBy.set(self.ds.getParameter(ParameterCodes.CAM_T_SCOLLNAME))
            self._idBarCode.set(self.ds.getParameter(ParameterCodes.CAM_T_SIDENT))
            self._qcBatchNum.set(self.ds.getParameter(ParameterCodes.CAM_T_SBATCHID))
            self._analysisRequest.set(self.ds.getParameter(ParameterCodes.CAM_T_SDESC4))
            if self.ds.getParameter(ParameterCodes.CAM_T_SSPBSTR1) in self._lablist.Laboratories:
                self._cb_labname.current(self._lablist.Laboratories.index(
                    self.ds.getParameter(ParameterCodes.CAM_T_SSPBSTR1)))
            if self.ds.getParameter(ParameterCodes.CAM_L_SSPI1) > 0:
                self._cb_wetdry.current(self.ds.getParameter(ParameterCodes.CAM_L_SSPI1) - 1)
        except Exception as exception:
            messagebox.showerror("Error In Script", traceback.format_exc())
            raise exception

    def _saveparams(self):
        """
        push all the custom parameters from the interface into the cam file
        this does not save the cam file, only pushes values into parameters
        this does not push the lab qualifier parameters, which is done both in
        "_onload" and in the "onClick"
        :return: None
        """
        self.ds.setParameter(ParameterCodes.CAM_T_SCOLLNAME, self._reportedBy.get())
        self.ds.setParameter(ParameterCodes.CAM_T_SIDENT, self._idBarCode.get())
        self.ds.setParameter(ParameterCodes.CAM_T_SBATCHID, self._qcBatchNum.get())
        self.ds.setParameter(ParameterCodes.CAM_T_SSPBSTR1, self._cb_labname.get())
        self.ds.setParameter(ParameterCodes.CAM_L_SSPI1, self._cb_wetdry.current() + 1)
        self.ds.setParameter(ParameterCodes.CAM_T_SDESC4, self._analysisRequest.get())

    def _cancel(self):
        # inform that this did not export anything.
        self._close()
        pass

    def _close(self):
        self._root.destroy()

    def _browse(self):
        filetypes = (
            ('csv files', '*.csv'),
            ('All files', '*.*')
        )
        fname = filedialog.askopenfilename(filetypes=filetypes)
        if fname:
            self._outfile.set(fname)

    def _show(self):
        self._root = Tk()
        self._root.title("Required information")
        # main window information
        mainframe = ttk.Frame(self._root, padding="3 3 12 12")
        mainframe.grid(column=0, row=0, sticky='N,W,E,S')
        self._root.columnconfigure(0, weight=1)
        self._root.rowconfigure(0, weight=1)

        # output file name and path
        self._outfile: StringVar = StringVar()
        ttk.Label(mainframe, text="Output File Name and Path").grid(column=1, row=0, sticky='W')
        otf_entry = ttk.Entry(mainframe, textvariable=self._outfile)
        otf_entry.grid(column=2, columnspan=3, row=0, sticky='E,W')
        # browse button for output file path
        otf_button = ttk.Button(mainframe, text="Browse", command=self._browse)
        otf_button.grid(column=5, row=0, sticky='w')

        # reported by label and entery field
        self._reportedBy = StringVar()
        ttk.Label(mainframe, text="Reported By").grid(column=1, row=1, sticky='E')
        rpt_entry = ttk.Entry(mainframe, width=20, textvariable=self._reportedBy)
        rpt_entry.grid(column=2, row=1, sticky='w,e')

        self._idBarCode = StringVar()
        ttk.Label(mainframe, text='ID/Barcode').grid(column=1, row=2, sticky='E')
        ibc_entry = ttk.Entry(mainframe, width=20, textvariable=self._idBarCode)
        ibc_entry.grid(column=2, row=2, sticky='w,e')

        self._qcBatchNum = StringVar()
        ttk.Label(mainframe, text='QC Batch #').grid(column=1, row=3, sticky='E')
        qbn_entry = ttk.Entry(mainframe, width=20, textvariable=self._qcBatchNum)
        qbn_entry.grid(column=2, row=3, sticky='w,e')

        self._analysisRequest = StringVar()
        ttk.Label(mainframe, text='Analysis Request').grid(column=1, row=4, sticky='E')
        qbn_entry = ttk.Entry(mainframe, width=20, textvariable=self._analysisRequest)
        qbn_entry.grid(column=2, row=4, sticky='w,e')

        # labName combobox
        ttk.Label(mainframe, text="Laboratory Name").grid(column=3, row=1, sticky='E')
        self._cb_labname = ttk.Combobox(mainframe)
        self._cb_labname.grid(column=4, row=1)
        self._cb_labname['values'] = self._lablist.Laboratories

        # Wet or Dry
        ttk.Label(mainframe, text="Wet or Dry").grid(column=1, row=5, sticky='E')
        self._cb_wetdry = ttk.Combobox(mainframe)
        self._cb_wetdry.grid(column=2, row=5)
        self._cb_wetdry['values'] = ['wet', 'dry']

        # Analysis Methodology
        default_analysis_value = "Gamma Spectroscopy"
        ttk.Label(mainframe, text="Analysis Methodology").grid(column=3, row=2, sticky='E')
        al = self._lablist.AnalysisMethodology
        startingindex = 0
        if default_analysis_value in al:
            startingindex = al.index(default_analysis_value)
        else:
            al.append(default_analysis_value)
            startingindex = al.index(default_analysis_value)
        self._cb_method = ttk.Combobox(mainframe, values=al)
        self._cb_method.grid(column=4, row=2)
        # has default value
        self._cb_method.current(startingindex)

        # list all nuclides from cam file and show in window for adding lab qualifiers
        def _popupwindow():
            listToShow = ["{0} - {1}".format(nuc.name,
                                             self._lablist.LabQualifiers[
                                                 int(self.ds.getParameter(
                                                     ParameterCodes.CAM_F_NCLSP1,
                                                     self._nucIndex.index(nuc.name) + 1)) - 1])
                          for nuc in self._nucIdList]

            def onClick():
                current_selections = [NuclideListbox.get(x).split(" - ")[0] for x in NuclideListbox.curselection()]
                for cur_nuc in current_selections:
                    self.ds.setParameter(ParameterCodes.CAM_F_NCLSP1, dp.current() + 1,
                                         self._nucIndex.index(cur_nuc.strip()) + 1)
                listToShow = ["{0} - {1}".format(nuc.name,
                                                 self._lablist.LabQualifiers[
                                                     int(self.ds.getParameter(
                                                         ParameterCodes.CAM_F_NCLSP1,
                                                         self._nucIndex.index(nuc.name) + 1)) - 1])
                              for nuc in self._nucIdList]
                NuclideListbox.configure(listvariable=StringVar(value=listToShow))

            nucLabQual = Toplevel(mainframe)
            nucLabQual.transient(mainframe)
            nucLabQual.overrideredirect(True)
            self.m_x = 0
            self.m_y = 0
            def mouseclick(e):
                print("coordinates: %s:%s" % (e.x, e.y))
                self.m_x = e.x
                self.m_y = e.y
            def movebar(e):
                nucLabQual.geometry(f'+{e.x_root - self.m_x}+{e.y_root - self.m_y}')

            nucLabQual.master.tk.eval(f'tk::PlaceWindow {str(nucLabQual)} center')
            nucLabQual.padding = "3 3 12 12"
            nucLabQual.title("Edit Lab Qualifiers")
            choices = StringVar(value=listToShow)

            title_bar = ttk.Label(nucLabQual, text="Edit Lab Qualifiers", justify="center", background='white',
                                  borderwidth=2, relief='flat')
            title_bar.grid(column=0, columnspan=5, row=0, sticky='e,w')
            title_bar.bind("<B1-Motion>", movebar)
            title_bar.bind("<ButtonPress-1>", mouseclick)

            #ttk.Separator(nucLabQual, orient="horizontal").grid(column=0, columnspan=5, row=1, sticky='e,w')
            NuclideListbox = Listbox(nucLabQual, listvariable=choices, selectmode="extended", exportselection=False,
                                     width=28)

            NuclideListbox.grid(column=1, columnspan=4, row=2, rowspan=3)

            dp = ttk.Combobox(nucLabQual, values=self._lablist.LabQualifiers, width=25)
            dp.grid(column=1, columnspan=2, row=5)
            ttk.Button(nucLabQual, text="Update Selected", command=onClick).grid(column=1, row=6)
            ttk.Button(nucLabQual, text="Done", command=nucLabQual.destroy).grid(column=2, row=6)
            for child in nucLabQual.winfo_children():
                if child != title_bar:
                    child.grid_configure(padx=5, pady=5)

        self._whatactreport = IntVar()
        whatActLableFrame = ttk.LabelFrame(mainframe, text="Report Activity Results Using")
        whatActLableFrame.grid(column=3, columnspan=2, row=3, rowspan=2)

        R1 = Radiobutton(whatActLableFrame, text="Wt Mean Act.", variable=self._whatactreport, value=1)
        R1.grid(column=0, row=0)  # , sticky='w')
        R1.select()
        R2 = Radiobutton(whatActLableFrame, text="Key Line", variable=self._whatactreport, value=2)
        R2.grid(column=1, row=0)  # , sticky='w')
        ttk.Button(mainframe, text="Edit Lab Qualifiers", command=_popupwindow).grid(column=3, columnspan=2, row=5)

        ttk.Separator(mainframe, orient="horizontal").grid(column=1, columnspan=5, row=6, sticky='e,w')
        ttk.Button(mainframe, text="Save & Export", command=self._pushtocam).grid(column=4, row=7)
        ttk.Button(mainframe, text="Cancel", command=self._cancel).grid(column=2, row=7)

        for child in mainframe.winfo_children():
            child.grid_configure(padx=5, pady=5)

        rpt_entry.focus()
        self._root.eval('tk::PlaceWindow . center')
        self._root.iconbitmap('C:/genie2k/app.ico')
        self._onLoad()
        self._loadparams()
        self._root.mainloop()


def setup(chan, step):
    try:
        ods = Datasource(chan)
        DisplayCustomPrompt(ods)
    except Exception as exception:
        messagebox.showerror("Error In Script", traceback.format_exc())
        raise exception


def execute(chan, step):
    try:
        ods = Datasource(chan)
        DisplayCustomPrompt(ods)
    except Exception as exception:
        messagebox.showerror("Error In Script", traceback.format_exc())
        raise exception
    return 0

# with Datasource() as ds:
#     # fn = r'C:/GENIE2K/CAMFILES/NBSSTD.CNF'
#     fn = r'C:/GENIE2K/vdmtemp/tmpfile - copy.CNF'
#     ds.open(fn, OpenOptions.READWRITE)
#     desc = ds.getProperty('descriptor')
#     execute(desc, 0)
